//
// Created by Attila Haluska on 8/1/18.
//

#include "lexer.h"

#include <string.h>

#include "compiler/lexer/token.h"
#include "memory_management/memory.h"

typedef struct keyword_t {
    const char *name;
    size_t length;
    TokenType type;
} Keyword;

static const Keyword keywords[] = {
        {"class",  5, TOKEN_CLASS},
        {"else",   4, TOKEN_ELSE},
        {"false",  5, TOKEN_FALSE},
        {"for",    3, TOKEN_FOR},
        {"func",   4, TOKEN_FUNC},
        {"if",     2, TOKEN_IF},
        {"nil",    3, TOKEN_NIL},
        {"print",  5, TOKEN_PRINT},
        {"return", 6, TOKEN_RETURN},
        {"super",  5, TOKEN_SUPER},
        {"this",   4, TOKEN_THIS},
        {"true",   4, TOKEN_TRUE},
        {"var",    3, TOKEN_VAR},
        {"while",  5, TOKEN_WHILE},
        {NULL,     0, TOKEN_EOF}
};

static Token *create_token(Lexer *lexer, TokenType type) {
    Token *token = NEW(Token);

    token->type = type;
    token->start = lexer->token_start;
    token->length = lexer->current_position - lexer->token_start;
    token->line = lexer->line;
    token->line_start = lexer->current_line_start;
    token->error_message = NULL;

    return token;
}

static Token *create_error_token(Lexer *lexer, const char *message) {
    Token *token = NEW(Token);

    token->type = TOKEN_ERROR;
    token->start = lexer->token_start;
    token->length = lexer->current_position - lexer->token_start;
    token->line = lexer->line;
    token->line_start = lexer->current_line_start;
    token->error_message = message;

    return token;
}

static bool is_at_end(Lexer *lexer) {
    return *lexer->current_position == '\0';
}

static char peek(Lexer *lexer) {
    return *lexer->current_position;
}

static char peek_next(Lexer *lexer) {
    if (is_at_end(lexer)) {
        return '\0';
    }

    return *(lexer->current_position + 1);
}

static char advance(Lexer *lexer) {
    lexer->current_position++;
    return *(lexer->current_position - 1);
}

static bool match(Lexer *lexer, char c) {
    if (is_at_end(lexer)) {
        return false;
    }

    if (*lexer->current_position != c) {
        return false;
    }

    lexer->current_position++;

    return true;
}

static bool is_digit(char c) {
    return c >= '0' && c <= '9';
}

static bool is_alpha(char c) {
    return (c >= 'a' && c <= 'z') ||
           (c >= 'A' && c <= 'Z') ||
           c == '_';
}

static bool is_alphanumeric(char c) {
    return is_alpha(c) || is_digit(c);
}

static bool is_whitespace(Lexer *lexer) {
    char c = peek(lexer);
    return c == ' ' || c == '\r' || c == '\t' || c == '\n';
}

static bool is_comment(Lexer *lexer) {
    char current = peek(lexer);
    char next = peek_next(lexer);
    return current == '/' && (next == '/' || next == '*');
}

static Token *scan_string(Lexer *lexer) {
    while (peek(lexer) != '"' && !is_at_end(lexer)) {
        char c = advance(lexer);

        if (c == '\n') {
            lexer->line++;
            lexer->current_line_start = lexer->current_position;
        }
    }

    if (is_at_end(lexer)) {
        return create_error_token(lexer, "Unterminated string");
    }

    advance(lexer);

    return create_token(lexer, TOKEN_STRING);
}

static Token *scan_number(Lexer *lexer) {
    while (is_digit(peek(lexer))) {
        advance(lexer);
    }

    if (peek(lexer) == '.' && is_digit(peek_next(lexer))) {
        advance(lexer);

        while (is_digit(peek(lexer))) {
            advance(lexer);
        }
    }

    return create_token(lexer, TOKEN_NUMBER);
}

static Token *scan_identifier(Lexer *lexer) {
    while (is_alphanumeric(peek(lexer))) {
        advance(lexer);
    }

    TokenType type = TOKEN_IDENTIFIER;

    size_t length = lexer->current_position - lexer->token_start;

    for (const Keyword *keyword = keywords; keyword->name != NULL; keyword++) {
        if (length == keyword->length && memcmp(lexer->token_start, keyword->name, length) == 0) {
            type = keyword->type;
            break;
        }
    }

    return create_token(lexer, type);
}

static Token *skip_comment(Lexer *lexer) {
    if (peek(lexer) == '/' && peek_next(lexer) == '/') {
        while (peek(lexer) != '\n' && !is_at_end(lexer)) {
            advance(lexer);
        }

        if (peek(lexer) == '\n') {
            lexer->line++;
            advance(lexer);
        }
    } else if (peek(lexer) == '/' && peek_next(lexer) == '*') {
        bool finished = false;

        while (!finished) {
            char c = advance(lexer);

            if (is_at_end(lexer)) {
                return create_error_token(lexer, "Unterminated comment");
            }

            if (c == '\n') {
                lexer->line++;
                lexer->current_line_start = lexer->current_position;
            }

            if (peek(lexer) == '*' && peek_next(lexer) == '/') {
                advance(lexer);
                advance(lexer);
                finished = true;
            }
        }
    }

    return NULL;
}

static void skip_whitespace(Lexer *lexer) {
    for (;;) {
        char c = peek(lexer);

        switch (c) {
            case ' ':
            case '\r':
            case '\t':
                advance(lexer);
                break;
            case '\n':
                advance(lexer);
                lexer->line++;
                lexer->current_line_start = lexer->current_position;
                break;
            default:
                return;
        }
    }
}

void lexer_init(Lexer *lexer, const char *source) {
    lexer->token_start = source;
    lexer->current_position = source;
    lexer->current_line_start = source;
    lexer->line = 1;
}

void lexer_free(Lexer *lexer) {
    lexer_init(lexer, NULL);
}

Token *lexer_next_token(Lexer *lexer) {
    bool comment;
    bool whitespace;

    do {
        skip_whitespace(lexer);
        Token *error = skip_comment(lexer);

        if (error != NULL) {
            return error;
        }

        comment = is_comment(lexer);
        whitespace = is_whitespace(lexer);
    } while (comment || whitespace);

    lexer->token_start = lexer->current_position;

    if (is_at_end(lexer)) {
        return create_token(lexer, TOKEN_EOF);
    }

    char c = advance(lexer);

    switch (c) {
        case '(':
            return create_token(lexer, TOKEN_LEFT_PAREN);
        case ')':
            return create_token(lexer, TOKEN_RIGHT_PAREN);
        case '{':
            return create_token(lexer, TOKEN_LEFT_BRACE);
        case '}':
            return create_token(lexer, TOKEN_RIGHT_BRACE);
        case ';':
            return create_token(lexer, TOKEN_SEMICOLON);
        case ',':
            return create_token(lexer, TOKEN_COMMA);
        case '.':
            return create_token(lexer, TOKEN_DOT);
        case '-':
            if (match(lexer, '-')) {
                return create_token(lexer, TOKEN_MINUS_MINUS);
            }

            return create_token(lexer, TOKEN_MINUS);
        case '+':
            if (match(lexer, '+')) {
                return create_token(lexer, TOKEN_PLUS_PLUS);
            }

            return create_token(lexer, TOKEN_PLUS);
        case '/': {
            return create_token(lexer, TOKEN_SLASH);
        }
        case '*':
            return create_token(lexer, TOKEN_STAR);
        case '?':
            return create_token(lexer, TOKEN_QUESTION);
        case ':':
            return create_token(lexer, TOKEN_COLON);
        case '!':
            if (match(lexer, '=')) {
                return create_token(lexer, TOKEN_BANG_EQUAL);
            }

            return create_token(lexer, TOKEN_BANG);
        case '=':
            if (match(lexer, '=')) {
                return create_token(lexer, TOKEN_EQUAL_EQUAL);
            }

            return create_token(lexer, TOKEN_EQUAL);
        case '<':
            if (match(lexer, '=')) {
                return create_token(lexer, TOKEN_LESS_EQUAL);
            }

            return create_token(lexer, TOKEN_LESS);
        case '>':
            if (match(lexer, '=')) {
                return create_token(lexer, TOKEN_GREATER_EQUAL);
            }

            return create_token(lexer, TOKEN_GREATER);
        case '&':
            if (match(lexer, '&')) {
                return create_token(lexer, TOKEN_AND_AND);
            }
        case '|':
            if (match(lexer, '|')) {
                return create_token(lexer, TOKEN_PIPE_PIPE);
            }
        case '"':
            return scan_string(lexer);
        default:
            if (is_digit(c)) {
                return scan_number(lexer);
            } else if (is_alpha(c)) {
                return scan_identifier(lexer);
            }

            break;
    }

    return create_error_token(lexer, "Unexpected character");
}
